options {
        language="Cpp";
}

{
#include "semantics.hh"
}
class ChaosParser extends Parser;
options {
        k=4;
}

chaos: (atom)+;

atom: (("'" fid:ID) {
                dynamic_cast<Chaos*>(this)->ChFn(fid->getText().c_str());
        }
        | ("$" vid:ID) {
                dynamic_cast<Chaos*>(this)->ChVar(vid->getText().c_str());
        }
        | str:STRING {
                dynamic_cast<Chaos*>(this)->ChString(str->getText().c_str());
        }
        | num:NUM {
                dynamic_cast<Chaos*>(this)->ChDigit(num->getText().c_str());
        }
        | ("precision" p:NUM) {
                dynamic_cast<Chaos*>(this)->ChPrecision(p->getText().c_str());
        }

        | "type" {
                dynamic_cast<Chaos*>(this)->ChType();
        }
        | "seq" {
                dynamic_cast<Chaos*>(this)->ChSequence();
        }
        | "?seq" {
                dynamic_cast<Chaos*>(this)->ChIsSequence();
        }
        | "len" {
                dynamic_cast<Chaos*>(this)->ChLength();
        }
        | "cat" {
		dynamic_cast<Chaos*>(this)->ChConcat();
	}
        | "slice" {
		dynamic_cast<Chaos*>(this)->ChSlice();
	}
        | "el" {
		dynamic_cast<Chaos*>(this)->ChElement();
	}
        | "search" {
		dynamic_cast<Chaos*>(this)->ChSearch();
	}
        | "rrepl" {
		dynamic_cast<Chaos*>(this)->ChRReplace();
	}
        | "repl" {
		dynamic_cast<Chaos*>(this)->ChReplace();
	}
        | "srepl" {
		dynamic_cast<Chaos*>(this)->ChSReplace();
	}
        | "ins" {
		dynamic_cast<Chaos*>(this)->ChInsert();
	}
        | "rdel" {
		dynamic_cast<Chaos*>(this)->ChRDel();
	}
        | "del" {
		dynamic_cast<Chaos*>(this)->ChDel();
	}
        | "iter" {
		dynamic_cast<Chaos*>(this)->ChIterator();
	}
        | "?iter" {
		dynamic_cast<Chaos*>(this)->ChIsIterator();
	}
        | "curr" {
		dynamic_cast<Chaos*>(this)->ChCurrent();
	}
        | "more" {
		dynamic_cast<Chaos*>(this)->ChMore();
	}
        | "next" {
		dynamic_cast<Chaos*>(this)->ChNext();
	}
        | "call" {
		dynamic_cast<Chaos*>(this)->ChCall();
	}
        | "true" {
		dynamic_cast<Chaos*>(this)->ChTrue();
	}
        | "set" {
		dynamic_cast<Chaos*>(this)->ChSetVar();
	}
        | "scope" {
		dynamic_cast<Chaos*>(this)->ChScope();
	}
        | "if" {
		dynamic_cast<Chaos*>(this)->ChIf();
	}
        | "while" {
		dynamic_cast<Chaos*>(this)->ChWhile();
	}
        | "eval" {
		dynamic_cast<Chaos*>(this)->ChEval();
	}
        | "dup" {
		dynamic_cast<Chaos*>(this)->ChDup();
	}
        | "drop" {
		dynamic_cast<Chaos*>(this)->ChDrop();
	}
        | "swap" {
		dynamic_cast<Chaos*>(this)->ChSwap();
	}
        | "." {
		dynamic_cast<Chaos*>(this)->ChPrint();
	}
        | "exit" {
		dynamic_cast<Chaos*>(this)->ChExit();
	}
        | "view" {
		dynamic_cast<Chaos*>(this)->ChView();
	}
        | "dump" {
		dynamic_cast<Chaos*>(this)->ChDump();
	}
        | "new" {
		dynamic_cast<Chaos*>(this)->ChNew();
	}
        | "names" {
		dynamic_cast<Chaos*>(this)->ChNames();
	}
        | "this" {
		dynamic_cast<Chaos*>(this)->ChThis();
	}
        | "built" {
		dynamic_cast<Chaos*>(this)->ChBuiltins();
	}

        | "==" {
		dynamic_cast<Chaos*>(this)->ChOpEq();
	}
        | "!=" {
		dynamic_cast<Chaos*>(this)->ChOpNEq();
	}
        | ">=" {
		dynamic_cast<Chaos*>(this)->ChOpGEq();
	}
        | "<=" {
		dynamic_cast<Chaos*>(this)->ChOpLEq();
	}
        | ">" {
		dynamic_cast<Chaos*>(this)->ChOpGT();
	}
        | "<" {
		dynamic_cast<Chaos*>(this)->ChOpLT();
	}

        | "pos" {
		dynamic_cast<Chaos*>(this)->ChPositive();
	}
        | "neg" {
		dynamic_cast<Chaos*>(this)->ChNegative();
	}

        | "+" {
		dynamic_cast<Chaos*>(this)->ChSum();
	}
        | "-" {
		dynamic_cast<Chaos*>(this)->ChSub();
	}
        | "*" {
		dynamic_cast<Chaos*>(this)->ChMult();
	}
        | "/" {
		dynamic_cast<Chaos*>(this)->ChDiv();
	}
        | "%" {
		dynamic_cast<Chaos*>(this)->ChMod();
	}
        | "~" {
		dynamic_cast<Chaos*>(this)->ChInvert();
	}
        | "&" {
		dynamic_cast<Chaos*>(this)->ChAnd();
	}
        | "^" {
		dynamic_cast<Chaos*>(this)->ChXor();
	}
        | "|" {
		dynamic_cast<Chaos*>(this)->ChOr();
	}
        | "<<" {
		dynamic_cast<Chaos*>(this)->ChLShift();
	}
        | ">>" {
		dynamic_cast<Chaos*>(this)->ChRShift();
	}
      );

class ChaosLexer extends Lexer;
options {
        charVocabulary = '\3'..'\377';
        k=4;
        testLiterals=true;
}

WS_ :	( ' ' | '\t' | '\n' { newline(); } | '\r' )+
        { $setType(Token::SKIP); }
        ;

protected LETTER: ('a'..'z' | 'A'..'Z' | '_');
protected DIGIT:  ('0'..'9');
protected DIGITS: (DIGIT)+;
protected HEX:	("0x" |"0X") (('0'..'9') | ('a'..'f') | ('A'..'F'))+;

NUM:    (HEX | DIGITS);

ID:	LETTER (LETTER | DIGIT)*;

protected QSTR:	'{' ( QSTR | ~'}' )* '}'; 
protected STR:	'"' (~'"')* '"';

STRING: STR | QSTR;
