/* forth.cc - simple forth interpreter as a level 1 test
Copyright (C) 1998,1999 Colin McCormack, Jordan B. Baker
see LICENSE (MD5 d7a11a4148f3858e544bf89cdde91727) for terms
*/
%{
#include "semantics.hh"
#define DEBUGLOG 1
        yy_buffer_state *yyFlexLexer::push_buffer(char *str, int len) {
            yy_buffer_state *old = yy_current_buffer;
            yy_buffer_state *b = yy_create_buffer(NULL, len);
            memcpy(b->yy_ch_buf, str, len);
            b->yy_ch_buf[len] = b->yy_ch_buf[len+1] = YY_END_OF_BUFFER_CHAR;
            b->yy_n_chars = len;	// we have len chars

            b->yy_fill_buffer = 0;	// don't fill it when it completes

            yy_switch_to_buffer(b);
	    yy_did_buffer_switch_on_eof = 0;
            return old;
        }

        void yyFlexLexer::pop_buffer(yy_buffer_state *old) {
            yy_delete_buffer(YY_CURRENT_BUFFER);
            yy_switch_to_buffer(old);
        }
%}
%option yyclass="Chaos"
%x str
DIGIT	[0-9]
HEX	0[xX][0-9a-fA-F]+
VAR	[$][a-zA-Z][a-zA-Z0-9_]*
STR	([^ \t\n{$'\0][^ \t\n\0]+)
QSTR	["]([^"]|(\\["]))*["]
ID	[a-zA-Z][a-zA-Z0-9_]*
OBRACE	[{]
CBRACE	[}]
%%

{DIGIT}+[.]{DIGIT}[Ee]?{DIGIT}* {
    ChReal(YYText());
}

{DIGIT}+ {
    ChDigit(YYText());
}

{HEX} {
    ChDigit(YYText());
}

{OBRACE} {
    if (!string) {
       string = (char*)malloc(1024);
       strl = 1024;
    }
    sp = string;
    paren = 1;
    BEGIN(str);
}

<str>{OBRACE} {
    paren++;
    if ((sp - string) >= strl) {
       sp -= (int)string;
       string = (char*)realloc(string, strl += 1024);
       sp += (int)string;
    }
    *sp++ = '{';
}

<str>{CBRACE} {
    if (!--paren) {
       *sp = '\0';
       ChBString(string);
       BEGIN(INITIAL);
    } else {
       if ((sp - string) >= strl) {
          sp -= (int)string;
          string = (char*)realloc(string, strl += 1024);
          sp += (int)string;
       }
       *sp++ = '}';
    }
}

<str>. {
    if ((sp - string) >= strl) {
       sp -= (int)string;
       string = (char*)realloc(string, strl += 1024);
       sp += (int)string;
    }
    *sp++ = YYText()[0];
}

"$"{ID} {
    ChVar(YYText());
}

"'pos"[a-zA-Z]* {
  ChPositive();
}

"'neg"[a-zA-Z]* {
  ChNegative();
}

"'eq"[a-zA-Z]* {
  ChOpEq();
}

"'add" {
  ChSum();
}

"+" {
  ChSum();
}

"'sub"[a-zA-Z]* {
  ChSub();
}

"-" {
  ChSub();
}

"'mul"[a-zA-Z]* {
  ChMult();
}

"*" {
  ChMult();
}

"'div[a-zA-Z]" {
  ChDiv();
}

"/" {
  ChDiv();
}

"'mod[a-zA-Z]" {
  ChDiv();
}

"%" {
  ChDiv();
}

"'invert" {
  ChInvert();
}

"~" {
  ChInvert();
}

"'and" {
  ChAnd();
}

"&" {
  ChAnd();
}

"'xor" {
  ChXor();
}

"^" {
  ChXor();
}

"'or" {
  ChOr();
}

"|" {
  ChOr();
}

"'lshift" {
  ChLShift();
}

"<<" {
  ChLShift();
}

"'rshift" {
  ChRShift();
}

">>" {
  ChRShift();
}

"'type" {
  ChType();
}

"'sequence" {
  ChSequence();
}

"'is"[Ss]eq[a-zA-Z]* {
  ChIsSequence();
}

"'len"[a-zA-Z]* {
  ChLength();
}

"'con"[a-zA-Z]* {
  ChConcat();
}

"'sli"[a-zA-Z]* {
  ChSlice();
}

"'el"[a-zA-Z]* {
  ChElement();
}

"'sea"[a-zA-Z]* {
  ChSearch();
}

"'rrepl"[a-zA-Z]* {
  ChRReplace();
}

"'repl"[a-zA-Z]* {
  ChReplace();
}

"'srepl"[a-zA-Z]* {
  ChSReplace();
}

"'ins"[a-zA-Z]* {
  ChInsert();
}

"'rdel"[a-zA-Z]* {
  ChRDel();
}

"'del"[a-zA-Z]* {
  ChDel();
}

"'iter"[a-zA-Z]* {
  ChIterator();
}

"'is"[iI]t[a-zA-Z]* {
  ChIsIterator();
}

"'cur"[a-zA-Z]* {
  ChCurrent();
}

"'more" {
  ChMore();
}

"'next" {
  ChNext();
}

"'call" {
  ChCall();
}

"==" {
  ChOpEq();
}

"!=" {
  ChOpNEq();
}

">=" {
  ChOpGEq();
}

">" {
  ChOpGT();
}

"<=" {
  ChOpLEq();
}

"<" {
  ChOpLT();
}

"'tru"[a-zA-Z]* {
  ChTrue();
}

"'set" {
  ChSetVar();
}

"'scope" {
    ChScope();
}

"'if" {
  ChIf();
}

"'while" {
  ChWhile();
}

"'eval" {
  ChEval();
}

"'dup" {
  ChDup();
}

"'drop" {
  ChDrop();
}

"'swap" {
  ChSwap();
}

[.] {
  ChPrint();
}

"'exit" {
  yyterminate();
}

"'view" {
  ChView();
}

"'dump" {
  ChDump();
}

"'new" {
  ChNew();
}

"'names" {
  ChNames();
}

"'th"[A-Za-z]* {
  ChThis();
}

"'bui"[A-Za-z]* {
  ChBuiltins();
}

"'precision "{DIGIT}+ {
    ChPrecision(YYText());
}

"'"{ID} {
  ChFn(YYText());
}

{QSTR} {
    ChString(YYText());
}

[ \t\n]+    /* eat up whitespace */  

<<EOF>> {
	yyterminate();
}

"\0" {
    /* I have no idea why this keeps getting NULs */
}

{STR} {
    ChBString(YYText());
}

.           cerr << "Unrecognized character: " << YYText() << '\n';

%%
