options {
        language="Cpp";
}

class ChaosParser extends Parser;
options {
}

chaos: 
        WS? (atom WS?)+;
atom: (
        (FN id:ID) {ChFn(id);}
        | (VAR id:ID) {ChVar(id);}
        | str:STRING {ChString(str);}
        | num:NUM {ChDigit(num);}
        | (PREC p:NUM} {ChPrecision(p);}

        | TYPE {ChType();}
        | SEQ {ChSequence();}
        | ISS {ChIsSequence();}
        | LEN {ChLength();}
        | CON {ChConcat();}
        | SLI {ChSlice();}
        | EL {ChElement();}
        | SEA {ChSearch();}
        | RREP {ChRReplace();}
        | REPL {ChReplace();}
        | SREP {ChSReplace();}
        | INS {ChInsert();}
        | RDEL {ChRDel();}
        | DEL {ChDel();}
        | ITER {ChIterator();}
        | ISIT {ChIsIterator();}
        | CURR {ChCurrent();}
        | MORE {ChMore();}
        | NEXT {ChNext();}
        | CALL {ChCall();}
        | TRUE {ChTrue();}
        | SET {ChSetVar();}
        | SCOPE {ChScope();}
        | IF {ChIf();}
        | WHILE {ChWhile();}
        | EVAL {ChEval();}
        | DUP {ChDup();}
        | DROP {ChDrop();}
        | SWAP {ChSwap();}
        | PRINT {ChPrint();}
        | EXIT {ChExit();}
        | VIEW {ChView();}
        | DUMP {ChDump();}
        | NEW {ChNew();}
        | NAMES {ChNames();}
        | THIS {ChThis();}
        | BUILT {ChBuiltins();}

        | {EQ | EQ1} {ChOpEq();}
        | NEQ {ChOpNEq();}
        | GEQ {ChOpGEq();}
        | LEQ {ChOpLEq();}
        | GT {ChOpGT();}
        | LT {ChOpLT();}

        | POS {ChPositive();}
        | NEG {ChNegative();}

        | ADD {ChSum();}
        | SUB {ChSub();}
        | MUL {ChMult();}
        | DIV {ChDiv();}
        | MOD {ChMod();}
        | INV {ChInvert();}
        | AND {ChAnd();}
        | XOR {ChXor();}
        | OR {ChOr();}
        | LSH {ChLShift();}
        | RSH {ChRShift();}
      )

class CalcLexer extends Lexer;
options {
        charVocabulary = '\3'..'\377';
}

tokens {
        TYPE="type";
        SEQ="seq";
        ISS="?seq";
        LEN="len";
        CON="cat";
        SLI="slice";
        EL="el";
        SEA="search";
        RREP="rrepl";
        REPL="repl";
        SREP="srepl";
        INS="ins";
        RDEL="rdel";
        DEL="del";
        ITER="iter";
        ISIT="?iter";
        CURR="curr";
        MORE="more";
        NEXT="next";
        CALL="call";
        TRUE="true";
        SET="set";
        SCOPE="scope";
        IF="if";
        WHILE="while";
        EVAL="eval";
        DUP="dup";
        DROP="drop";
        SWAP="swap";
        PRINT=".";
        EXIT="exit";
        VIEW="view";
        DUMP="dump";
        NEW="new";
        NAMES="names";
        THIS="this";
        BUILT="built";
        PREC="precision";
        
        EQ="eq";
        EQ1="==";
        NEQ="!=";
        GEQ=">=";
        LEQ="<=";
        GT=">";
        LT="<";

        POS="pos";
        NEG="neg";

        ADD="+";
        SUB="-";
        MUL="*";
        DIV="/";
        MOD="%";
        INV="~';
        AND="&";
        XOR="^";
        OR=""|";
        LSH="<<";
        RSH=">>";
}

WS :	( ' ' | '\t' | '\n' { newline(); } | '\r' )+
        { $setType(Token.SKIP); }
        ;

LETTER: ('a'..'z' | 'A'..'Z' | '_');
DIGIT:  ('0'..'9');
DIGITS: DIGIT+;
HEX:	'0' ('x'|'X')(0..9 | a..f | A..F)+
NUM:    DIGITS | HEX;

ID	options { testLiterals=true; }
        : LETTER (LETTER | DIGIT)*;

QSTR:	'{' ( QSTR | ~'}' )* '}'; 

FN:	'\'' ID;

VAR:	'$' ID;

STR:	'"' (~'"' | '\''"')* '"';
STRING: '"' (~'"")* '"';

STRING: STR | QSTR;
