// semantics of Chaos
// Copyright (C) 1998,1999 Colin McCormack, Jordan B. Baker
// see LICENSE (MD5 d7a11a4148f3858e544bf89cdde91727) for terms

static char *id __attribute__((unused))="$Id$";
//#define DEBUGLOG 1
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <Data.hh>
#include <Store.hh>
#include <Tuple.hh>
#include <Segment.hh>
#include <String.hh>
#include <List.hh>
#include <Dict.hh>
#include <Symbol.hh>
#include <Error.hh>
#include <Builtin.hh>
#include <BigInt.hh>
#include <Real.hh>
#include <elf.hh>
#include "semantics.hh"

// #define this as 1 to get lex logging
#define FDEBUG 0
void onError();

# define strdupa(s)                                                           \
  (__extension__                                                              \
    ({                                                                        \
      __const char *__old = (s);                                              \
      size_t __len = strlen (__old) + 1;                                      \
      char *__new = (char *)__builtin_alloca (__len);                                 \
      (char *) memcpy (__new, __old, __len);                                  \
    }))

void Chaos::ChPush(Slot el)
{
    DEBLOG(el->dump(cerr) << " Push / " << (1 + length()) << '\n');
    push(el);
}

Slot Chaos::ChPop()
{
    Slot el = pop();
    DEBLOG(el->dump(cerr) << " Pop / " << length() <<'\n');
    return el;
}


void Chaos::ChVar(const char *var)
{
    DEBLOG(cerr << "Var\n");
    char *sdup = strdupa(var);
    Slot sym = store->Names->search(sdup+1);
#ifdef DEBUGLOG
    if (sym) {
        DEBLOG(cerr << "Var: " << var << " " << sym->typeId());
        if (sym->typeEQ(store->typeSymbol)) {
            cerr << " = " ; ((Symbol*)sym)->value()->dump(cerr);
        }
        cerr << '\n';
    } else {
        DEBLOG(cerr << "Var: " << var << " NULL\n");
    }
#endif
    if (!sym) {
        throw new Error("dictnf", var, "key not found");
    }
    ChPush(((Symbol*)sym)->value());
}

void Chaos::ChSetVar()
{
    Slot val = ChPop();
    Slot name = ChPop();
    DEBLOG(cerr << "SetVar " << (char*)name
           << " = " << (char*)val
           << '\n');    
    store->Names->insert(name, val);
}

void Chaos::ChType()
{
    DEBLOG(cerr << "Type\n");
    Slot arg = ChPop();
    ChPush(arg.typeId());
}

void Chaos::ChNew()
{
    DEBLOG(cerr << "New\n");
    char name[256];
    Slot tname = ChPop();
    Slot args = ChPop();
    
    sprintf(name, "__%d%sRC4Slot", strlen(tname), (char*)tname);
    cerr << "Looking for Builtin " << name << '\n';
    Builtin *b = ((Symbol*)(store->Registered->slice(name)))->value();
    if (b) {
        // found our type
        b->call(args);
    } else {
        throw new Error("type", tname, "no such type known");
    }
}

void Chaos::ChDigit(const char *digits)
{
    DEBLOG(cerr << "Digit\n");
    long n = strtol(digits, NULL, 0);
    if (errno == ERANGE) {
        // value's too big
        Slot num(new BigInt(digits));
        ChPush(num);
    } else {
        Slot num(n);
        ChPush(num);
    }
//  fprintf(stderr, "List len is %d\n", stack->length());
}

void Chaos::ChPrecision(const char *digits)
{
    DEBLOG(cerr << "Precision\n");
    unsigned long n = strtol(index(digits, ' ')+1, NULL, 0);
    cerr << "precision `" << digits << "` / " << n << '\n';
    if (errno == ERANGE) {
        throw new Error("range", digits, "too long");
    } else {
        Real::precision(n);
    }
//  fprintf(stderr, "List len is %d\n", stack->length());
}

void Chaos::ChReal(const char *digits)
{
    DEBLOG(cerr << "Real\n");
    Slot num(new Real(digits));
    ChPush(num);
}

void Chaos::ChString(const char *string)
{
    DEBLOG(cerr << "String\n");
    char *sdup = strdupa(string);
    sdup[strlen(sdup)-1] = '\0';
    Slot str(sdup+1);
    ChPush(str);
}

void Chaos::ChBString(const char *string)
{
    DEBLOG(cerr << "BString:" << string << '\n');
    if (string[0] == '\0') {
        onError();
    }
    Slot str(string);
    ChPush(str);
}

void Chaos::ChEval()
{
    DEBLOG(cerr << "Eval\n");
    Slot buffer = ChPop();

    if (buffer && buffer.isString()) {
#ifdef DEBUGLOG
        cerr << "EVAL: " << buffer->typeId()
             << ' ' << buffer;
        if (buffer.isString()) {
            cerr << ' ' << *(String*)buffer;
        }
        cerr << " len: " << buffer->length();
        cerr << '\n';
        cerr << "Stack: " << length() << '\n';
#endif
        yy_buffer_state *oldstate
            = push_buffer((char*)buffer, buffer->length());
        while(1) {
            try {
                int done = yylex();
                if (!done)
                    break;
            }
            catch (Error *e) {
                e->dump(cerr) << '\n';
                break;
                // cerr << "stack underflow\n";
            } catch (...) {
                cerr << "THROW\n";
                break;
            }
        }
        pop_buffer(oldstate);
    } else {
        if (buffer)
            ChPush(buffer);
#ifdef DEBUGLOG
        if (buffer) {
            cerr << "nonstrEVAL: " << buffer->typeId()
                 << ' ' << buffer << '\n';
        } else {
            cerr << "Eval: NULL\n";
        }
        cerr << "Stack: " << length() << '\n';
#endif
    }
}

void Chaos::ChIf()
{
    DEBLOG(cerr << "If\n");
    //cout << "If: "; ChView();
    if (ChPop()->truth()) {
        ChDrop();
    } else {
        ChSwap();
        ChDrop();
    }
    ChEval();
}

void Chaos::ChWhile()
{
    DEBLOG(cerr << "While\n");
    Slot test = ChPop();
    Slot body = ChPop();
    Slot tvar;
    DEBLOG(cerr << test->typeId() << " " << test->dump(cerr) << " test0\n");
    ChPush(test);
    ChEval();
    tvar = ChPop();
    DEBLOG(cerr << tvar->typeId() << " " << tvar->dump(cerr) << " val\n");
    while (tvar->truth()) {
        DEBLOG(cerr << body->typeId() << " " << body->dump(cerr) << " body\n");
        ChPush(body);
        ChEval();
        DEBLOG(cerr << test->typeId() << " " << test->dump(cerr) << " test\n");
        ChPush(test);
        ChEval();
        tvar = ChPop();
        DEBLOG(cerr << tvar->typeId() << " " << tvar->dump(cerr) << " val\n");
    }
}

void Chaos::ChTrue()
{
    DEBLOG(cerr << "True\n");
    ChPush(ChPop()->truth());
}

void Chaos::ChSum()
{
    DEBLOG(cerr << "Sum\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->add(arg2));
}

void Chaos::ChSub()
{
    DEBLOG(cerr << "Sub\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->subtract(arg2));
}

void Chaos::ChMult()
{
    DEBLOG(cerr << "Mult\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->multiply(arg2));
}

void Chaos::ChDiv()
{
    DEBLOG(cerr << "Div\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->divide(arg2));
}

void Chaos::ChDup()
{
    DEBLOG(cerr << "Dup\n");
    Slot arg1 = ChPop();
    ChPush(arg1);
    ChPush(arg1);
}

void Chaos::ChSwap()
{
    DEBLOG(cerr << "Swap\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1);
    ChPush(arg2);
}

void Chaos::ChDrop()
{
    DEBLOG(cerr << "Drop\n");
    ChPop();
}

void Chaos::ChPrint()
{
    DEBLOG(cerr << "Print\n");
    Slot arg1 = ChPop();
    arg1->dump(cout) << '\n';
    //<< " (" << sizeof(typeof(*(Data*)arg1)) << ")\n";
}

void Chaos::ChView()
{
    DEBLOG(cerr << "View\n");
    //    dump(cout) << '\n';

    for (int len = Length() - 1; len >= 0; len --) {
        slice(len)->dump(cout) << "\n";
    }
}

void Chaos::ChDump()
{
    DEBLOG(cerr << "Dump\n");
    dump(cout) << '\n';
}

void Chaos::ChNames()
{
    DEBLOG(cerr << "Names\n");
    ChPush(store->Names);
}

void Chaos::ChThis()
{
    DEBLOG(cerr << "This\n");
    ChPush(clone());
}

void Chaos::ChBuiltins()
{
    DEBLOG(cerr << "Builtins\n");
    store->Registered->dump(cout) << '\n';
}

void Chaos::ChFn(const char *name)
{
    DEBLOG(cerr << "Fn\n");
    ChVar(name);
    ChEval();
}

void Chaos::ChScope()
{
    DEBLOG(cerr << "Scope\n");
    char *sdup = ChPop();
    ChVar(sdup);

    Chaos *lexer = new Chaos(ChPop());
    while(1) {
        try {
            if (!lexer->yylex())
                break;
        }
        catch (Error *e) {
            e->dump(cerr) << '\n';
            break;
            // cerr << "stack underflow\n";
        }
    }
    ChPush(lexer);
}

void Chaos::ChPositive()
{
    DEBLOG(cerr << "Positive\n");
    Slot arg1 = ChPop();
    ChPush(arg1->positive());
}

void Chaos::ChNegative()
{
    DEBLOG(cerr << "Negative\n");
    Slot arg1 = ChPop();
    ChPush(arg1->negative());
}

void Chaos::ChInvert()
{
    DEBLOG(cerr << "Invert\n");
    Slot arg1 = ChPop();
    ChPush(arg1->invert());
}

void Chaos::ChAnd()
{
    DEBLOG(cerr << "And\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->and(arg2));
}

void Chaos::ChXor()
{
    DEBLOG(cerr << "Xor\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->xor(arg2));
}

void Chaos::ChOr()
{
    DEBLOG(cerr << "Or\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->or(arg2));
}

void Chaos::ChLShift()
{
    DEBLOG(cerr << "LShift\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->lshift(arg2));
}

void Chaos::ChRShift()
{
    DEBLOG(cerr << "RShift\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->rshift(arg2));
}

void Chaos::ChSequence()
{
    DEBLOG(cerr << "Sequence\n");
    Slot arg1 = ChPop();
    ChPush(arg1->toSequence());
}

void Chaos::ChIsSequence()
{
    DEBLOG(cerr << "IsSequence\n");
    Slot arg1 = ChPop();
    ChPush(arg1->isSequence());
}

void Chaos::ChLength()
{
    DEBLOG(cerr << "Length\n");
    Slot arg1 = ChPop();
    ChPush(arg1->length());
}

void Chaos::ChConcat()
{
    DEBLOG(cerr << "Concat\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->concat(arg2));
}

void Chaos::ChSlice()
{
    DEBLOG(cerr << "Slice\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    Slot arg3 = ChPop();
    ChPush(arg1->slice(arg2, arg3));
}

void Chaos::ChElement()
{
    DEBLOG(cerr << "Element\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->slice(arg2));
}

void Chaos::ChSearch()
{
    DEBLOG(cerr << "Search\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->search(arg2));
}

void Chaos::ChRReplace()
{
    DEBLOG(cerr << "RReplace\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    Slot arg3 = ChPop();
    Slot arg4 = ChPop();
    ChPush(arg1->replace(arg2, arg3, arg4));
}

void Chaos::ChReplace()
{
    DEBLOG(cerr << "Replace\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    Slot arg3 = ChPop();
    ChPush(arg1->replace(arg2, arg3));
}

void Chaos::ChSReplace()
{
    DEBLOG(cerr << "SReplace\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->replace(arg2));
}

void Chaos::ChInsert()
{
    DEBLOG(cerr << "Insert\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    Slot arg3 = ChPop();
    ChPush(arg1->insert(arg2, arg3));
}

void Chaos::ChRDel()
{
    DEBLOG(cerr << "RDel\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    Slot arg3 = ChPop();
    ChPush(arg1->del(arg2, arg3));
}

void Chaos::ChDel()
{
    DEBLOG(cerr << "Del\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->del(arg2));
}

void Chaos::ChIterator()
{
    DEBLOG(cerr << "Iterator\n");
    Slot arg1 = ChPop();
    ChPush(arg1->iterator());
}

void Chaos::ChIsIterator()
{
    DEBLOG(cerr << "IsIterator\n");
    Slot arg1 = ChPop();
    ChPush(arg1->isIterator());
}

void Chaos::ChCurrent()
{
    DEBLOG(cerr << "Current\n");
    Slot arg1 = ChPop();
    ChPush(arg1->Current());
}

void Chaos::ChMore()
{
    DEBLOG(cerr << "More\n");
    Slot arg1 = ChPop();
    ChPush(arg1->More());
}

void Chaos::ChNext()
{
    DEBLOG(cerr << "Next\n");
    Slot arg1 = ChPop();
    ChPush(arg1->Next());
}

void Chaos::ChCall()
{
    DEBLOG(cerr << "Call\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1->call(arg2));
}

void Chaos::ChOpEq()
{
    DEBLOG(cerr << "OpEq\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1 == arg2);
}

void Chaos::ChOpNEq()
{
    DEBLOG(cerr << "OpNEq\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg1 != arg2);
}

void Chaos::ChOpGEq()
{
    DEBLOG(cerr << "OpGEq\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg2 >= arg1);
}

void Chaos::ChOpGT()
{
    DEBLOG(cerr << "OpGT\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg2 > arg1);
}

void Chaos::ChOpLEq()
{
    DEBLOG(cerr << "OpLEq\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg2 <= arg1);
}

void Chaos::ChOpLT()
{
    DEBLOG(cerr << "OpLT\n");
    Slot arg1 = ChPop();
    Slot arg2 = ChPop();
    ChPush(arg2 < arg1);
}

Chaos::Chaos(const Slot& init)
    : string(NULL),
      strl(0),
      sp(string),
      paren(0)
{
    set_debug(FDEBUG);
    DEBLOG(cerr << "Chaos" << this
         << " from Slot " << init.typeId()
         << '\n');
    push_buffer((char*)(String*)init, init->length());
}

Chaos::Chaos(istream *infile)
    : yyFlexLexer(infile),
      string(NULL),
      strl(0),
      sp(string),
      paren(0)
{
    set_debug(FDEBUG);
    DEBLOG(cerr << "Chaos" << this
           << " from stdin "
           << '\n');
}

Chaos::~Chaos()
{}

void Chaos::reset()
{
}

void ChExit()
{
}

#if 0
int Chaos::LexerInput(char* buf, int max_size)
{
    if (buffer && buffer.isString()) {
        // String driven
        DEBLOG(cerr << buffer->typeId() << " " << buffer << '\n');
        DEBLOG(buffer->dump(cerr) << " GOT " << buffer << '\n');
        int size = max_size <? (buffer->length() - offset);
        memcpy(buf, offset + (char*)*(String*)buffer, size);
        offset += size;
        return size;
    } else {
        // stream driven
        return yyFlexLexer::LexerInput(buf, max_size);
    }
}
#endif
#if 0
String *Chaos::operator += (String *b)
{
    cerr << "BINGO\n";
    if (buffer) {
        buffer = buffer->concat(b);
    } else {
        buffer = b;
    }
    return buffer;
}
#endif

extern void onError();
/** called when Elf has loaded this module
 */
void initElf(Elf *elf)
{
    //cerr << "chaos's initElf called\n";

    Chaos *lexer;
    Slot sym;

    // register our Builtins with ColdStore
    try {
        RegisterElf(elf);
    } catch (Error *e) {
        e->dump(cerr) << '\n';
        throw e;
    }

    sym = store->Names->search("Chaos");
    if (sym) {
        lexer = (Chaos*)(Data*)((Symbol *)sym)->value();
        DEBLOG(cerr << "Chaos " << lexer->typeId()
               << ' ' << lexer
               << '\n');
        lexer->reset();
    } else {
        lexer = new Chaos();
        DEBLOG(cerr << "Chaos " << lexer->typeId()
               << ' ' << lexer
               << '\n');
        store->Names->insert("Chaos", lexer);
    }

    //lexer->dump(cerr) << '\n';
    //Memory::Assert();

    //yyin = stdin;
    while(1) {
        try {
            if (!lexer->yylex()) {
                break;
            }
        }
        catch (Error *e) {
            e->dump(cerr) << '\n';
            //lexer->ChPush((Symbol*)*e);
            // cerr << "stack underflow\n";
        } catch (char *err) {
            cerr << "ERROR: " << err << '\n';
            onError();
        }
    }
    onError();
    //delete lexer;
}

int Chaos::yywrap()
{
    return 1;
}

extern "C" {
    int yywrap()
    {
        return 1;
    }
}
