// Data - class from which all ColdStore entities must derive
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms

static char *id __attribute__((unused))="$Id$";

#include <ctype.h>
#include <typeinfo>
#include <iostream.h>

#include "Data.hh"
#include "Store.hh"
#include "Builtin.hh"
#include "Task.hh"

#include "String.hh"
#include "Tuple.hh"
#include "Error.hh"
#include "List.hh"

//static Data *dummy = (Data*)0;

Data::Data()
{}

Data::~Data()
{
    // if we're destroying something that's not unreferenced, it's an error
    assert(refcount() == 0);
}

// returns a mutable copy of this
Data *Data::mutate(void *) const
{
    return (Data*)this;		// by default - objects are mutable
}

// returns a mutable copy of this
Slot Data::identity(void *) const
{
    return this;		// by default - objects are mutable
}

void Data::check(int) const
{
    validGuard();
    //unimpl("No check operator defined");
}

// type - return the address of this type's RTTI method
// it's assumed this address will be unique, and will uniquely
// characterise the instance.
void *Data::type() const
{
    assert("NULL passed to Data::type()");
    //    throw new Error("illegal", NULL, "NULL passed to Data::type()");
    return this->_vptr;
}

const char *Data::typeId() const {
    return typeid(*this).name();
}

bool Data::typeEQ(const Builtin *m) const
{
    void **modrtti = (void **)(m->vtbl()); // pointer to RTTI function

    DEBLOG(cerr << "typeEQ: type()=" << type()
           << ", modrtti=" << modrtti
           << '\n');
    return type() == modrtti;
}

Slot Data::unimpl(const char *message) const
{
    throw new Error("unimplemented", this, message);
    return this;
}

Slot Data::Unimpl(const char *message)
{
    throw new Error("unimplemented", (Data*)NULL, message);
    return (Data*)NULL;
}

// the object's truth value
bool Data::truth() const
{
    return true;	// Slot will convert this to an encoded Integer
}

ostream &Data::dump(ostream& out) const
{
    if (this != (Data*)0) {
        toconstruct()->dump(out);
    } else {
        out << "NULL";
    }
    return out;
}

//  constructor args to recreate object
Slot Data::toconstruct() const
{
    return unimpl("No constructor rendering defined");
}

//  1,0,-1 depending on order
int Data::order(const Slot &arg) const
{
    return (Data*)this - (Data *)arg;	// default strong equality
  // Note: Slot will have converted this to an encoded Integer
}

//  equality predicate
bool Data::equal(const Slot &arg) const
{
    return order(arg) == 0;	// default is strong equality
}

//  1,0,-1 depending on order of Type
int Data::typeOrder(const Slot &arg) const
{
    // compare the object type
    return (int)type() - (int)(arg->type());
    //  return strcmp(typeid(this).name(), typeid((Data*)arg).name());
}

//  unary '+', absolute value
Slot Data::positive()
{
    return unimpl("No monadic `+' operator defined");
}

//  unary '-', negative absolute value
Slot Data::negative()
{
    return unimpl("No monadic `-' operator defined");
}

//  dyadic `+', add
Slot Data::add(const Slot &arg)
{
    return unimpl("No dyadic `+' operator defined");
}

//  dyadic `-', subtract
Slot Data::subtract(const Slot &arg)
{
    return unimpl("No dyadic `-' operator defined");
}

//  dyadic `*', multiply
Slot Data::multiply(const Slot &arg)
{
    return unimpl("No dyadic `*' operator defined");
}

//  dyadic '/', divide
Slot Data::divide(const Slot &arg)
{
    return unimpl("No dyadic `/' operator defined");
}

//  dyadic '%', modulo
Slot Data::modulo(const Slot &arg)
{
    return unimpl("No dyadic `%' operator defined");
}

// bitwise
//  unary '~', invert
Slot Data::invert()
{
    return unimpl("No monadic `~' operator defined");
}

//  dyadic '&', bitwise and
Slot Data::and(const Slot &arg)
{
    return unimpl("No dyadic `&' operator defined");
}

//  dyadic '^', bitwise xor
Slot Data::xor(const Slot &arg)
{
    return unimpl("No dyadic `^' operator defined");
}

//  dyadic '|', bitwise or
Slot Data::or(const Slot &arg)
{
    return unimpl("No dyadic `|' operator defined");
}

//  dyadic '<<', left shift
Slot Data::lshift(const Slot &arg)
{
    return unimpl("No dyadic `<<' operator defined");
}

//  dyadic '>>', right shift
Slot Data::rshift(const Slot &arg)
{
    return unimpl("No dyadic `>>' operator defined");
}

// predicate - is object a sequence
bool Data::isSequence()
{
    return false;
}

//  length as sequence
int Data::length() const
{
    return unimpl("No monadic `length' operator defined");
}

//  concatenate two sequences
Slot Data::concat(const Slot &arg)
{
    return unimpl("No dyadic `concat' operator defined");
}

//  dyadic `[]', index
Slot Data::slice(const Slot &index) const
{
    // return unimpl("No dyadic `concat' operator defined");
    return slice(index, (Data*)0);
}

//  dyadic `[]', index
Slot Data::slice(const Slot &from, const Slot &len) const
{
    return unimpl("No dyadic `[]' operator defined");
}

//  dyadic `in', matching subrange
Slot Data::search(const Slot &search) const
{
    return unimpl("No dyadic `search' operator defined");
}

Slot Data::del(Slot const &what)
{
    return unimpl("No diadic `del' operator defined");
}

Slot Data::del(Slot const &from, Slot const &len)
{
    return unimpl("No triadic `del' operator defined");
}

Slot Data::insert(Slot const &before, Slot const &value)
{
    return unimpl("No triadic `insert' operator defined");
}

Slot Data::insert(Slot const &value)
{
    return unimpl("No dyadic `insert' operator defined");
}

//  replace subrange
Slot Data::replace(const Slot &from, const Slot &len, const Slot &value)
{
    return unimpl("No quadradic `[]=' operator defined");
}

//  replace item
Slot Data::replace(const Slot &from, const Slot &value)
{
    return unimpl("No triadic `[]=' operator defined");
}

//  replace
Slot Data::replace(const Slot &value)
{
    return unimpl("No dyadic `[]=' operator defined");
}

//  return an iterator
Slot Data::iterator() const
{
    return unimpl("No iterator operator defined");
}

// return an iterable Sequence as a List
Slot Data::toSequence() const
{
    // has to be done the with an iterator
    Slot result = new List(length());
    Slot it = iterator();
    
    while (it->More()) {
        result->concat(it->Current());
        it = it->Next();
    }
    
    return result;
}

// predicate - is object an iterator?
bool Data::isIterator()
{
    return false;
}


Slot Data::Current()
{
    return unimpl("No Current operator defined");
}

bool Data::More() const
{
    unimpl("No More operator defined");
    return false;
}

Slot Data::Next()
{
    return unimpl("No Next operator defined");
}

//  call object (Method, Builtin)
Slot Data::call(const Slot&) const
{
    return unimpl("No call operator defined");
}
