// Pool - maintain an allocation pool of templated type
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms
// 	$Id $

#ifndef POOL_HH
#define POOL_HH

#include <Q.hh>
#include <qvmm.h>

template <class T>
class Pool
    : public Memory,
      public Q< Pool<T> >
{
    static Qh<Pool<T> > all;	// pointer to Q of all Pools of T
    struct PoolCell {
        PoolCell next;		// pointer to next free PoolCell
        T *freeH;		// free list for this Pool
        PoolCell()
            : T
    }
    static const size_t def_size = (Mmap::page_size / sizeof(T*))-1;
    T *content[def_size];

    T *get();

protected:
    ~Pool();

    friend T;
    static T* Get();
    void release(T *el);

public:
    Pool<T>();
};

template <class T>
Pool<T>::Pool()
    : Q<Pool<T> >(all)
{
    DEBLOG(cerr << "Pool new " << def_size << '\n');
    for (size_t i = 0; i < def_size; i++) {
        content[i] = new T((int)(content i + 1), this);
    }
    (T*&)(content[def_size - 1]) = NULL;
    freeH = content[0];
}

template <class T>
Pool<T>::~Pool()
{
    throw "Trying to delete Pool";
}

template <class T>
T *Pool<T>::get()
{
    DEBLOG(cerr << "Pool " << this
           << '(' << next() << ',' << prev() << ") get ");
    if (freeH) {
        // we have a free item
        T *gotit = freeH;
        freeH = *(T**)freeH;
        gotit->_home = this;
        DEBLOG(cerr << gotit << '\n');
        return gotit;
    }
    DEBLOG(cerr << "NULL\n");
    return NULL;
}

template <class T>
T* Pool<T>::Get()
{
    if (!all) {
        Pool<T> *p =  new Pool<T> ();
        return p->get();            
    } else {
        // find a Pool with free
        for (Qi<Pool<T> > pi(all); pi; pi++) {
            T *found = ((Pool<T>*)pi)->get();
            if (found)
                return found;
        }
        // no Pool has free, create new Pool
        DEBLOG(cerr << " new Pool\n");
        Pool<T> *p = new Pool<T>();
        return p->get();
    }
}

template <class T>
void Pool<T>::release(T *el)
{
    DEBLOG(cerr << "Pool " << this << " release " << el << '\n');
    //el->~T();	// explicitly destroy element
    
    // add element to the free linked list
    *(T**)el = freeH;
    freeH = el;
}
#endif

