SHLIB_SUFFIX = .so

LOADLIBES = -rdynamic
CFLAGS = -Wall -Woverloaded-virtual \
	-I. -I../include -I../coldstore \
	-DSHLIB_EXT=\"${SHLIB_SUFFIX}\"

#	-fvtable-gc 

# Linker Values
#LOADLIBES += -Xlinker --dynamic-linker=`pwd`/../libs/ld-2.1.2.so

#LOADLIBES += -Xlinker -rpath -Xlinker `pwd`/../libs
#LOADLIBES += -Xlinker -rpath -Xlinker /usr/local/lib
#LOADLIBES += -Xlinker -rpath -Xlinker /usr/share/lib
LDLIBS = -L../libs ../libs/elf.so ../libs/libqvmm.so
#LDLIBS += -L /usr/share/lib 
LDLIBS += -ldl -lelf -lgmp2

#../coldstore/libcold.so

# optimisation
CFLAGS += -O2 

# debugging
LOADLIBES += -ggdb3
CFLAGS += -ggdb3

# DEBUGLOG controls extremely verbose DEBUG logging 
#CFLAGS += -DDEBUGLOG

# GUARD_ALLOC controls allocation guards and checking
CFLAGS += -DGUARD_ALLOC

# BTREE_INTEGRITY controls exhaustive testing of the entire heap (debug only)
#CFLAGS += -DBTREE_INTEGRITY

# The following makes all mallocs tram allocations
CFLAGS += -DWRAP_MALLOC
LDFLAGS += -Xlinker --wrap -Xlinker malloc -Xlinker --wrap -Xlinker free -Xlinker --wrap -Xlinker realloc

CXXFLAGS = ${CFLAGS} -frtti -fvtable-thunks -fno-implicit-templates
#CXXFLAGS += -Weffc++
SUFFIXES += .so

%.o : %.cc
	$(CXX) -c $(CPPFLAGS) -fPIC $(CXXFLAGS) $<

%.o : %.cpp
	$(CXX) -c $(CPPFLAGS) -fPIC $(CXXFLAGS) $<

%.so : %.o
	${CXX} ${CXXFLAGS} ${LDFLAGS} -shared -o $@ $^ ${SOLIBS}

%.so :
	${CXX} ${CXXFLAGS} ${LDFLAGS} -shared -o $@ $^ ${SOLIBS}

%: %.o
	$(CXX) $(LOADLIBES) ${LDFLAGS} $(LDLIBS) -o $@ $^

%.cpp: %.g
	CLASSPATH=/usr/share/antlr-2.6.0 java antlr.Tool $^

# auto-construct dependancies
%.d: %.cc
	$(SHELL) -ec '$(CXX) -MM $(CXXFLAGS) $< \
		| sed '\''s/$*\\.o[ :]*/& $@/g'\'' > $@'

%.d: %.cpp
	$(SHELL) -ec '$(CXX) -MM $(CXXFLAGS) $< \
		| sed '\''s/$*\\.o[ :]*/& $@/g'\'' > $@'

%.m:
	cd ../$(@:.m=); make all

all:

.PHONY: kdocs

kdocs:
	-mkdir kdoc
	-kdoc -d kdoc $^

.PHONY: clean
CLEAN = *.a *.so *.o *~ kdoc

clean:
	-rm -rf ${CLEAN}

distclean:
	-rm -rf *.a *.so *.o *~ kdoc *.d
