// Elf - interning of Elf DSOs
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms

static char *id __attribute__((unused))="$Id$";

// This file contains modified code (where marked <LIBELF>)
// from ftp://sunsite.unc.edu/pub/Linux/libs/libelf-0.6.4.tar.gz
// Clearly, my copyright does not apply to that code, nor does
// the author of that code (Michael "Tired" Riepe 
// <michael@stud.uni-hannover.de>) bear responsibility for my mods.

#include <iostream.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <elf.h>                // AKPM
#include <sys/elf.h>            // AKPM
#include <libelf.h>
//#include <bfd.h>

#include <qvmm.h>
#include "elf.hh"
#include "elfexe.hh"

template class Q<Elf>;
template class Qh<Elf>;
template class Q<Csym>;
template class Qh<Csym>;
ElfExe *Elf::exe = NULL;
//Q<Elf> Elf::all;	// header of all Elf Q
extern void bkpt();

Elf::Elf(const char *name)
    : Q<Elf>(all)
{
    strncpy(elf, name, 256);
#ifdef DEBUGLOG
    cerr << "new Elf(" << elf << "): " << this << " vptr=" << this->_vptr << '\n';
#endif
}

Elf::~Elf()
{
    throw "Can't destroy Elfs";
}


void Elf::reset()
{
    _so = NULL;	// NULL persistent .so handle refs
    for (Qi<Csym> sp(symbols); sp.More(); sp.Next()) {
        ((Csym*)sp)->reset();
    }
}


void Elf::initialize(ElfExe *_exe) {
    exe = _exe;
    if (all) {
        // reset all the Elf instances
        for (Qi<Elf> ep(all); ep.More(); ep.Next()) {
            return ((Elf*)ep)->reset();
        }
    }
}

Csym *Elf::findSym(const char *sym) const
{
    for (Qi<Csym> sp(symbols); sp.More(); sp.Next()) {
        if (!strcmp(((Csym*)sp)->name, sym)) {
#ifdef DEBUGLOG
            cerr << "findSym(" << sym << "): " << (Csym*)sp << '\n';
#endif
            return (Csym*)sp;
        }
    }
    //cerr << "findSym(" << sym << "): NULL\n";
    return NULL;
}

bool Elf::addSym(const char *sym)
{
    if (findSym(sym)) {
#ifdef DEBUGLOG
        cerr << "addSym(" << sym << "): false\n";
#endif
        return false;
    }

    csym(sym, Csym::UNKNOWN, 0);	// add a placeholder
    return true;
}

Csym *Elf::csym(const char *mangled, Csym::Type type, int size)
{
#ifdef DEBUGLOG
    cerr << "csym(" << mangled << ", " << size << ")\n";
#endif
    return new (size) Csym(this, mangled, type, size);
}

Elf *Elf::findElf(const char *elf)
{
    for (Qi<Elf> ep(all); ep.More(); ep.Next()) {
        if (!strcmp(((Elf*)ep)->elf, elf)) {
#ifdef DEBUGLOG
            cerr << "findElf(" << elf
                 << "): " << (Elf*)ep << '\n';
#endif
            return (Elf*)ep;
        }
    }
    return NULL;
}

void Elf::forallElf(void (fn)(Elf*))
{
    for (Qi<Elf> ep(all); ep.More(); ep.Next()) {
        (*fn)((Elf*)ep);
    }
}

void Elf::forallSym(void (fn)(Elf*,Csym*))
{
    for (Qi<Csym> sp(symbols); sp.More(); sp.Next()) {
        (*fn)(this, (Csym*)sp);
    }
}

// create new Csym
Csym *Elf::snarfSym(char *symname, Elf32_Sym *sym)
{
    Csym *fsym = NULL;
    switch (ELF32_ST_TYPE(sym->st_info)) {
    case STT_FUNC:
        fsym = csym(symname, Csym::FUNCTION,
                    sym->st_size);
        break;
    case STT_OBJECT:
        fsym = csym(symname, Csym::DATA,
                    sym->st_size);
        break;
    default:
        break;
    }
    return fsym;
}

// Code from Elk which snarfs symbol tables.
void Elf::snarf(Policy personality)
{
    Elf		*elf_ptr;
    Elf_Scn	*elf_scn_ptr = NULL, *symtab_scn_ptr = NULL;
    Elf_Data	*elf_data_ptr = NULL;
    Elf32_Ehdr	*elf_ehdr_ptr = NULL;
    Elf32_Shdr	*elf_shdr_ptr = NULL,
		*symtab_ptr = NULL;
    size_t	elf_str_index = 0;
    size_t	shstrndx;
    char	*section_name;

#ifdef DEBUGLOG
    cerr << "Elf::snarf " << elf << '\n';
#endif

    // open the .so file
    int	f = open(elf, O_RDONLY);

    if (f == -1)
        throw "Can't open Elf .so file";

    if (elf_version (EV_CURRENT) == EV_NONE) {
        ::close(f);
        throw "Elf .so file Elf version out of date";
    }
    if ((elf_ptr = elf_begin (f, ELF_C_READ, (Elf *)NULL)) == NULL) {
        ::close(f);
        throw "can't elf_begin() .so file";
    }

    // get the elf header, so we'll know where to look for the section 
    // names. <LIBELF>
    if ((elf_ehdr_ptr = elf32_getehdr (elf_ptr)) == NULL) {
        elf_end(elf_ptr);
        ::close(f);
	throw "no elf header in .so file";
    }
    shstrndx = elf_ehdr_ptr->e_shstrndx;

    // look for the symbol and string tables
    while ((elf_scn_ptr = elf_nextscn (elf_ptr, elf_scn_ptr))) {
	if ((elf_shdr_ptr = elf32_getshdr (elf_scn_ptr)) == NULL) {
            ::close(f);
            throw "can't get section header in .so file";
        }
	if (elf_shdr_ptr->sh_type == SHT_STRTAB) {
            // save the index to the string table for later use by 
            // elf_strptr().
	    section_name = elf_strptr (elf_ptr, shstrndx,
                                       (size_t)elf_shdr_ptr->sh_name);
	    if (strcmp (section_name, ".strtab") == 0 ||
                strcmp (section_name, ".dynstr") == 0) {
		elf_str_index = elf_ndxscn (elf_scn_ptr);
	    }
	}
	else if (elf_shdr_ptr->sh_type == SHT_SYMTAB ||
		 elf_shdr_ptr->sh_type == SHT_DYNSYM) {
	    symtab_ptr = elf_shdr_ptr;
	    symtab_scn_ptr = elf_scn_ptr;
	}
    }

    // check the .so
    if (!symtab_ptr) {
        elf_end(elf_ptr);
        ::close(f);
        throw "no symbol table in .so file";
    }
    if (!elf_str_index) {
        elf_end(elf_ptr);
        ::close(f);
        throw "no string table in .so file";
    }

    // we've located the symbol table -- go through it and save the names 
    // of the interesting symbols.
    while ((elf_data_ptr = elf_getdata(symtab_scn_ptr, elf_data_ptr))) {
	char *name = NULL;
	unsigned int symbol_count;
	Elf32_Sym *symbol_ptr = (Elf32_Sym*)elf_data_ptr->d_buf;
	Elf32_Sym *current_symbol;
        Csym *fsym;

	for (symbol_count = 1;
	     symbol_count < symtab_ptr->sh_size / symtab_ptr->sh_entsize;
	     symbol_count++) {
	    current_symbol = symbol_ptr + symbol_count;

	    if (ELF32_ST_BIND(current_symbol->st_info) != STB_GLOBAL
                && ELF32_ST_BIND(current_symbol->st_info) != STB_WEAK) {
                // discard non-globals
		continue;
	    }

            name = elf_strptr(elf_ptr, elf_str_index,
                              (size_t)current_symbol->st_name);
            //</LIBELF>
	    if (name == NULL) {
                elf_end(elf_ptr);
		::close(f);
		throw elf_errmsg(elf_errno());
	    }

            // see if this symbol is already interned
            fsym = findSym(name);
            if (fsym) {
                // existing interned symbol
                switch (fsym->_type) {
                case Csym::UNKNOWN: {
                    // We've located a symbol which was previously unknown
#ifdef DEBUGLOG
                    cerr << "found UNKNOWN symol "
                         << name << " as " << fsym << '\n';
#endif
                    // previously unknown symbol
                    delete fsym;	// destroy placeholder

                    // create new Csym
                    if (!snarfSym(name, current_symbol)) {
                        elf_end(elf_ptr);
                        ::close(f);
                        throw "Requested bad symbol type";
                    }
                    break;
                }
                case Csym::FUNCTION:
#ifdef DEBUGLOG
                    cerr << "found requested FUNCTION symol "
                         << name << " as " << fsym << '\n';
#endif
                    // nothing to be done
                    break;

                case Csym::DATA:
                case Csym::VTBL:
#ifdef DEBUGLOG
                    cerr << "found requested DATA/VTBL symol "
                         << name << " as " << fsym << '\n';
#endif
                    if (fsym->len() < current_symbol->st_size) {
                        elf_end(elf_ptr);
                        ::close(f);
                        throw "symbol changed size catastrophically";
                    }
                    break;
                default:
                    elf_end(elf_ptr);
                    ::close(f);
                    throw "corrupt Csym found";
                }
            } else if (!strncmp("__vt_", name, 5)) {
                // vtbls are interned regardless
#ifdef DEBUGLOG
                cerr << "found new VTBL symbol "
                     << name
                     << " st_info " << (void*)current_symbol->st_info
                     << " st_shndx " << current_symbol->st_shndx
                     << '\n';
#endif
                csym(name, Csym::VTBL, current_symbol->st_size);
            } else if (personality == COLDSTORE) {
                // coldstore constructors need to be interned too
                // A coldstore constructor is a symbol of the form
                // ending in 'RC4Slot' and beginning __nOBJNAME
                char *ending = "RC4Slot";
                if (!strncmp("__", name, 2)
                    && ((name[2] >= '0') && (name[2] <= '9'))
                    && !strcmp(ending, name + strlen(name) - strlen(ending))
                    && (ELF32_ST_TYPE(current_symbol->st_info) == STT_FUNC)) {
#ifdef DEBUGLOG
                    cerr << "found new Coldstore Constructor "
                         << name
                         << " st_info " << (void*)current_symbol->st_info
                         << " st_shndx " << current_symbol->st_shndx
                         << '\n';
#endif
                    csym(name, Csym::FUNCTION, current_symbol->st_size);
                }
            } else if (personality == STL) {
                // STL all globals need to be interned
                if (ELF32_ST_TYPE(current_symbol->st_info) == STT_FUNC) {
                    if (!snarfSym(name, current_symbol)) {
                        elf_end(elf_ptr);
                        ::close(f);
                        throw "Requested bad symbol type";
                    }
                }
            } else {
                // unwanted symbol - do nothing
            }
	}
    }
    elf_end(elf_ptr);
    ::close(f);
}

int Elf::shadow()
{
    int symcnt = 0;	// count of symbols defined

    // count up the symbols and check that they're all defined
    for (Qi<Csym> sp(symbols); sp.More(); sp.Next()) {
        switch (((Csym*)sp)->_type) {
        case Csym::FUNCTION:
        case Csym::DATA:
        case Csym::VTBL:
            symcnt++;
            continue;
        default:
            // let undefined symbols pass silently
            throw "undefined symbol in Elf";
        }
    }

#ifdef DEBUGLOG
    cerr << "Elf::shadow " << elf
         << ' ' << symcnt << " symbols \n";
#endif

    // Elf's Csyms into a bfd symbol table
    symcnt = 0;

    bkpt();
    for (Qi<Csym> sp(symbols); sp.More(); sp.Next()) {
        Csym *csym = (Csym*)sp;
        if (csym->type() != Csym::UNKNOWN) {
#ifdef DEBUGLOG
            cerr << "Elf::shadow "
                 << elf << " making sym "
                 << csym->name << " @ " << csym->address() << '\n';
#endif    
            exe->addSym(csym->symname(), csym->address(), csym->len(),
                   (csym->type() == Csym::FUNCTION)
                   ?STT_FUNC
                   :STT_OBJECT);
            symcnt++;
        }
    }

    return symcnt;
}

void Elf::load()
{
#ifdef DEBUGLOG
    cerr << "Elf::load \n";
#endif

    // load the .so file
    //_so = (link_map*)dlopen(elf, RTLD_GLOBAL | RTLD_LAZY);
    _so = (link_map*)dlopen(elf, RTLD_GLOBAL | RTLD_NOW);
    if (!_so) {
        throw dlerror();
    }
}

void Elf::resolve()
{
    for (Qi<Csym> sp(symbols); sp.More(); sp.Next()) {
        if (((Csym*)sp)->_type != Csym::UNKNOWN) {
            void *sym = dlsym(_so, ((Csym *)sp)->name);
            char *error = dlerror();
            if (error) {
                throw error;
            }

            // don't reloc a symbol twice
            if (!((Csym*)sp)->isLoaded()) {
                ((Csym*)sp)->reloc(sym);
            }
        }
    }
}

void Elf::intern(Policy policy)
{
    snarf(policy);
    shadow();
    load();
    resolve();
}

Csym::Csym(Elf *elf, const char *mangled, Csym::Type type, int len)
    : Q<Csym>(elf->symbols),
      in(elf),
      _type(type),
      _len(len),
      loaded(false)
{
    strncpy(name, mangled, 256);
#ifdef DEBUGLOG
    cerr << this
         << " new Csym " << symname()
         << " in: " << in
         << " type: " << type
         << " len " << len
         << '\n';
#endif
}

Csym::~Csym()
{
    throw "Can't destroy Csyms";
}

void Csym::reset()
{
    loaded = false;
}

bool Csym::isLoaded() const
{
    return loaded;
}

void Csym::reloc(void *so)
{
    switch (_type) {
    case FUNCTION:
        // form up a JMP instruction
#ifdef DEBUGLOG
        cerr << "Csym::reloc FUNCTION " << name << ' ' << so << '\n';
#endif
        content[0] = JMP;
        //*((int*)(content + 1)) = (int)so;
        *((int*)(content + 1)) = ((int)so) - (int)content - 5; // feh, Intel.
#ifdef DEBUGLOG
        cerr << "set " << (void*)content << " to " << JMP << ' ' << so << '\n';
#endif
        break;

    case DATA:
        // copy the content
#ifdef DEBUGLOG
        cerr << "Csym::reloc DATA "
             << name << ' ' << so << ' ' << _len
             << " from " << so << " to " << (void*)content << '\n';
#endif
        memcpy((void *)content, so, _len);
        break;

    case VTBL:
        // copy the content
#ifdef DEBUGLOG
        cerr << "Csym::reloc VTBL "
             << name << ' ' << so << ' ' << _len
             << " from " << so << " to " << (void*)content << '\n';
#endif
        memcpy((void *)content, so, _len);
        break;

    default:
        throw "Unknown type in Csym::reloc";
    }
    loaded = true;	// flag Csym as loaded
}

void *Elf::getSym(const char *name) const
{
    void *sym = dlsym(_so, name);
    char *error = dlerror();
    if (error)
        throw error;
    return sym;
}

void Elf::close()
{
    char *error;

    if (_so) {
        dlclose(_so);
        if ((error = dlerror()))
            throw error;
    }
}

/** default AutoLoad (ie: none)
 */
AutoLoad autoload_elfs[] = {
    {NULL}
};

extern void bkpt();

/** function to unload Elfs
 */
static void elf_shutdown(Elf *elf)
{
    elf->close();
}

void *InitElf::load(const char *name, int flags /*= RTLD_LAZY*/)
{
    void *so = dlopen(name, flags);
    if (!so) {
        throw dlerror();
    }
#ifdef DEBUGLOG
        cerr << "InitElf::load " << name << " -> " << so << '\n';
#endif
    return so;
}

void *InitElf::getSym(void *so, char *name, bool chuck)
{
    void *sym = dlsym(so, name);
    char *err = dlerror();
    if (err && chuck) {
        throw err;
    } else {
#ifdef DEBUGLOG
        cerr << "InitElf::getSym " << name << " -> " << sym << " from " << so << '\n';
#endif
    }
    return sym;
}

/** copy symbol contents from .so into the store.
 */
char *InitElf::copy_reloc(void *from, char *what, size_t len)
{
    void *symt = getSym(main_so, what);
    void *symf = getSym(from, what);
    
#ifdef DEBUGLOG
        cerr << "InitElf::copy_reloc " << symt
             << ", " << symf
             << " " << len
             << '\n';
#endif
    memcpy(symt, symf, len);
    return NULL;
}

Elf *this_elf;

void dumpObj(const link_map *);

InitElf::InitElf(const char *elflib)
    : exe(new ElfExe()), 
      main_so((link_map *)load(NULL, RTLD_GLOBAL | RTLD_NOW)),
      elf((link_map *)load(elflib, RTLD_GLOBAL | RTLD_NOW /*| RTLD_LAZY */))
{
    //link_map *qvmm = (link_map *)load("../libs/libqvmm.so", RTLD_GLOBAL | RTLD_NOW /*| RTLD_LAZY */);

#ifdef DEBUGLOG
    dumpObj(main_so);
    dumpObj(elf);
    //dumpObj(qvmm);
#endif
    //dlclose(elf);	// remove extra reference
    
    // copy relocate the global.so Elf structures
    copy_reloc(elf, "__vt_3Elf", 12);
    copy_reloc(elf, "__vt_4Csym", 12);
    copy_reloc(elf, "__vt_t1Q1Z3Elf", 12);
    copy_reloc(elf, "__vt_t1Q1Z4Csym", 12);
    //copy_reloc(qvmm, "__vt_8RefCount", 20);

    // initialize Elf module
    Elf::initialize(exe);
    
    // autoload requested 
    bkpt();	// call to bkpt
    for (AutoLoad *to_load = autoload_elfs; to_load->name; to_load++) {
        cerr << "Autoload " << to_load->name << '\n';
        Elf *elf = Elf::findElf(to_load->name);
        if (!elf)
            elf = new Elf(to_load->name);

        this_elf = elf;

        try {
            elf->intern(to_load->type);
        } catch (char *err) {
            cerr << "Error: " << err << '\n';
        }

        // now call the initElf routine
        {
            void (*initElf)(Elf*);

            if (!to_load->initFn)
                to_load->initFn = "initElf__FP3Elf";
            initElf = (void(*)(Elf*))getSym(elf->_so, to_load->initFn, false);
            if (initElf) {
                try {
                    bkpt();
                    (*initElf)(elf);
                } catch (char *message) {
                    cerr << "Autoload threw: "
                         << message
                         << '\n';
                } catch (...) {
                    cerr << "Autoload: "
                         << to_load->initFn
                         << " threw\n";
                }
            } else {
                cerr << "No initElf function for " << to_load->name << '\n';
            }
        }
    }
    cerr << "DONE\n";
    bkpt();	// call to bkpt
}

InitElf::~InitElf()
{
    Elf::forallElf(elf_shutdown);
    delete exe;

#ifdef DEBUGLOG
    cerr << "InitElf destroyed!!\n";
#endif
}

// Local Variables:
// mode:C++
// End:
