// Elf - dumping of Elf DSOs
// Copyright (C) 1998,1999 Colin McCormack,
// see LICENSE (MD5 f5220f8f599e5e926f37cf32efe3ab68) for terms
// 	$Id

void dumpSectName(const link_map *l, Elf32_Section s)
{
    switch (s) {
    case SHN_UNDEF:
        cerr << "*UND*"; break;
    case SHN_ABS:
        cerr << "*ABS*"; break;
    case SHN_COMMON:
        cerr << "*COM*"; break;
    default:
        if (s < SHN_LORESERVE) {
            cerr << s;
        } else {
            cerr << "*UNKNOWN*" << s;
        }
    }
}

void dumpBind(const link_map *l, int b)
{
    switch (b) {
    case STB_LOCAL:
        cerr << "LOCAL"; break;
    case STB_GLOBAL:
        cerr << "GLOBAL"; break;
    case STB_WEAK:
        cerr << "WEAK"; break;
    default:
        cerr << "UNKNOWN"; break;
    }
}

void dumpType(const link_map *l, int t)
{
    switch (t) {
    case STT_NOTYPE:
        cerr << "NOTYPE"; break;
    case STT_OBJECT:
        cerr << "OBJECT"; break;
    case STT_FUNC:
        cerr << "FUNC"; break;
    case STT_SECTION:
        cerr << "SECTION"; break;
    case STT_FILE:
        cerr << "FILE"; break;
    default:
        cerr << "UNKNOWN"; break;
    }
}

void dumpSym(const link_map *l, const ElfW(Sym) *sym)
{
    const char *strtab = (char *)(l->l_info[DT_STRTAB]->d_un.d_ptr);

    cerr << sym << ' ' << strtab + sym->st_name << ':';
    cerr << "\n\tvalue: " << (void*)(sym->st_value);
    cerr << "\n\tsize: " << sym->st_size;
    cerr << "\n\ttype: "; dumpType(l, ELF32_ST_TYPE(sym->st_info));
    cerr << "\n\tbind: "; dumpBind(l, ELF32_ST_BIND(sym->st_info));
    //cerr << "\n\tother: " << sym->st_other;
    cerr << "\n\tsection: "; dumpSectName(l, sym->st_shndx);
    cerr << '\n';
}

void dumpSymIdx(const link_map *l, ElfW(Symndx) symidx)
{
    const ElfW(Sym) *sym = (ElfW(Sym) *)(l->l_info[DT_SYMTAB]->d_un.d_ptr);
    sym = sym + symidx;

    // dump the symbol sym
    dumpSym(l, sym);
}

void dumpObj(const link_map *l)
{
    int *hash = (int *)(l->l_info[DT_HASH]->d_un.d_ptr + l->l_addr);
    cerr << l << ':';
    cerr << " l_next: " << l->l_next;
    cerr << " l_prev: " << l->l_prev;
    cerr << '\n';
    cerr << l->l_name;
    cerr << "\n\tl_type: " << l->l_type;
    cerr << "\n\tl_addr: " << (void*)(l->l_addr);
    cerr << "\n\tsymtab: " << (void *)(l->l_info[DT_SYMTAB]->d_un.d_ptr);
    cerr << "\n\thash: " << hash;
    cerr << "\tbuckets: " << *hash;
    cerr << "\tchains: " << hash[1];
    //cerr << "\tsyment: " << l->l_info[DT_SYMENT]->d_un.d_val;
    cerr << "\n\tl_nbuckets: " << l->l_nbuckets;
    cerr << "\tl_buckets: " << l->l_buckets;
    cerr << "\tl_chain: " << l->l_chain;
    cerr << "\n\tstrtab: " << (void*)(l->l_info[DT_STRTAB]->d_un.d_ptr);
    cerr << "\tstrtab size: " << l->l_info[DT_STRSZ]->d_un.d_val;
    cerr << "\n\tl_versyms: " << l->l_versyms;
    cerr << '\n';
}
